#!/usr/bin/env python
# -*- coding: utf-8 -*-
import zlib

import pendulum


class Alarmed(dict):
    """
    Alarm event model

    Attributes:
        dt_begin (pendulum.Datetime): begin
        dt_end (pendulum.Datetime): end
        code (str): Error code
        message (str): Error message
        alarm_id (str): Alarm ID

    >>> al = Alarmed()
    Traceback (most recent call last):
        ...
    ValueError: Empty Code
    >>> allen = Alarmed(code="x")
    Traceback (most recent call last):
        ...
    ValueError: No Begin
    """

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

        for r_key in ("dt_begin", "dt_end", "code", "message"):
            try:
                self[r_key]
            except KeyError:
                self[r_key] = None

        if self.get("dt") and {self.get("dt_begin"), self.get("dt_end")} == {
            None
        }:
            value = self["dt"]
            self["dt_begin"] = value
            self["dt_end"] = value
            del self["dt"]

        for dt_key in ("dt_begin", "dt_end"):
            if isinstance(self[dt_key], str):
                self[dt_key] = pendulum.parse(self[dt_key])

        if self.get("text") and self.get("message") is None:
            self["message"] = self["text"]
            del self["text"]

        self["alarm_id"] = self._mk_alarm_id()

        if self.dt_begin is None:
            raise ValueError("No Begin")

    @classmethod
    def mk_alarm_id(cls, code, message):
        """
        Generate an alarm ID for an alarm log item containing an alarm code and
        some descriptive text.

        If the text is empty the Alarm ID equals the alarm code.

        Alarm ID consist of the Alarm Code and an hex encoded CRC32 hash of the
        text concatenated by a dot.

        .. warning::

            It is assumed that the text is encoded in valid ``UTF-8``!

        Text normalisation:

            #. Leading and trailing whitespaces are removed
            #. String is converted to lowercase

        Args:
            code (str): Alarm code
            message (str): Alarm text

        Returns:
            str: Alarm ID

        >>> Alarmed.mk_alarm_id('710531', "710531 | HYDRAULIK: Bitte Öl auffüllen")
        '710531.5CB5A110'
        >>> Alarmed.mk_alarm_id('700244', '700244 | BAR FEEDER: Axes and spindles of machine blocked')
        '700244.B8D47740'
        >>> Alarmed.mk_alarm_id('27097', '27097 | NCU_1: SPL-Start nicht erfolgt')
        '27097.D020A11'

        """
        if not code:
            code = ""

        if not code:
            raise ValueError("Empty Code")

        if not message:
            return code

        mangled = message.strip().lower()
        crc = "{:X}".format(zlib.crc32(mangled.encode("utf-8")))

        return f"{code}.{crc}"

    def _mk_alarm_id(self):
        return Alarmed.mk_alarm_id(self.get("code"), self.get("message"))

    def __str__(self):
        period = pendulum.period(self.dt_begin, self.dt_end)

        return "<Alarm> {code:>12} @{dt_begin} {ts:16.1f}s: {message}".format(
            dt_begin=self.dt_begin.format("YYYY-MM-DD HH:mm:ss.S"),
            code=self.code,
            message=self.message,
            ts=period.total_seconds(),
        )

    @property
    def repeater_v2_record(self):
        """
        Get alarm data as a list to be used for repeater V2 endpoint.

        Returns:
            list: Alarm Repeater V2 record

        >>> alf = Alarmed(dt="2021-01-01T17:03:00Z", code="red", message="UH-OH!")
        >>> alf.repeater_v2_record
        ['2021-01-01T17:03:00+00:00', 'red.B3F10B52']
        >>> ale = Alarmed(dt="2021-01-01T17:03:00Z", code="red")
        >>> ale.repeater_v2_record
        ['2021-01-01T17:03:00+00:00', 'red']
        """
        return [self.dt_begin.to_rfc3339_string(), self.alarm_id]

    @property
    def dt_begin(self):
        return self["dt_begin"]

    @property
    def dt_end(self):
        return self["dt_end"]

    @property
    def code(self):
        return self["code"]

    @property
    def alarm_id(self):
        return self["alarm_id"]

    @property
    def message(self):
        return self["message"]


if __name__ == "__main__":
    import doctest

    (FAILED, SUCCEEDED) = doctest.testmod()
    print("[doctest] SUCCEEDED/FAILED: {:d}/{:d}".format(SUCCEEDED, FAILED))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
from io import BytesIO

import boto3
import botocore.exceptions
import orjson


class JSONDocumentOnS3Controller(object):
    """
    Controller class for reading and writing JSON encoded documents.
    """

    # TODO: change variable name
    env_bucket_key = "TANGIBLE_BUCKET"

    def __init__(self, *args, **kwargs):
        self.log = logging.getLogger(__name__)

        if kwargs.get("bucket_name") is None:
            self.bucket_name = os.environ.get(self.env_bucket_key)
        else:
            self.bucket_name = kwargs.get("bucket_name")

        self.region_name = kwargs.get("region_name")

        if self.bucket_name is None:
            mfg = (
                "No bucket name?! Please set either environment variable "
                "{!r} or provide "
                "*bucket_name* parameter".format(self.env_bucket_key)
            )
            self.log.error(mfg)

            raise ValueError(mfg)

    def s3_upload(self, key, value):
        """
        Upload JSON content to S3.

        Args:
            key: key
            value: value

        """
        s3 = boto3.client("s3", region_name=self.region_name)
        filelike = BytesIO()
        filelike.write(orjson.dumps(value))
        filelike.seek(0)

        s3.upload_fileobj(filelike, self.bucket_name, key)

    def s3_delete(self, key):
        s3 = boto3.client("s3", region_name=self.region_name)
        s3.delete_object(Key=key, Bucket=self.bucket_name)

    def s3_download(self, key):
        """
        Download JSON content from S3.

        Args:
            key:

        Returns:
            object: contents
        """
        s3 = boto3.client("s3", region_name=self.region_name)
        filelike = BytesIO()

        s3.download_fileobj(self.bucket_name, key, filelike)

        return orjson.loads(filelike.getvalue())

    def __delitem__(self, key):
        self.s3_delete(key)

    def __getitem__(self, key):
        try:
            return self.s3_download(key)
        except botocore.exceptions.ClientError as bexc:
            if bexc.response["Error"]["Code"] == "404":
                raise KeyError(key)
            response_id = bexc.response["ResponseMetadata"]["RequestId"]
            self.log.error(
                "Unexpected S3 error: {!s} Request ID: {!s}".format(bexc, response_id)
            )

    def __setitem__(self, key, value):
        self.s3_upload(key, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import copy

import pendulum
from coshed.tools import log_traceback

from coshed_model.documents import JSONDocumentOnS3Controller


class Cellule(list):
    """
    >>> c = Cellule("x", pendulum.datetime(2021, 8, 8))
    >>> c.value
    'x'
    >>> c.dt
    DateTime(2021, 8, 8, 0, 0, 0, tzinfo=Timezone('+00:00'))
    >>> import orjson
    >>> orjson.dumps(c)
    b'["2021-08-08T00:00:00+00:00","x"]'
    """

    def __init__(self, value, dt=None, **kwargs):
        if dt is None:
            dt = pendulum.now().to_rfc3339_string()

        if not isinstance(dt, str):
            dt_str = dt.to_rfc3339_string()
        else:
            dt_str = dt

        list.__init__(self, (dt_str, value))

    @property
    def value(self):
        return self[1]

    @property
    def dt(self):
        return pendulum.parse(self[0])

    def __str__(self) -> str:
        return "{!r} @{!s}".format(self.value, self.dt.format("YYYY-MM-DD HH:mm:ss Z"))


class CelluleBorked(ValueError):
    pass


class ValueOutdated(ValueError):
    pass


class PoupinCerveau(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.dirty = False

    def __setitem__(self, key, value):
        if key not in self and isinstance(value, list):
            if len(value) == 2:
                dataset = [Cellule(raw_val, raw_dt) for (raw_dt, raw_val) in value]

                return super().__setitem__(key, dataset)

        dt_now = pendulum.now()

        try:
            raw_dataset = self[key]
        except KeyError:
            raw_dataset = [
                (dt_now, None),
                (dt_now, None),
            ]

        dataset = [Cellule(raw_val, raw_dt) for (raw_dt, raw_val) in raw_dataset]

        previous = dataset[0]
        current = dataset[1]

        if not previous.dt <= current.dt:
            raise CelluleBorked()

        if not current.dt <= dt_now:
            raise ValueOutdated("Will not accept data prior to {!r}".format(current.dt))

        if current.value != value:
            previous = current
            current = Cellule(value, dt_now)
            # print("{!s} -> {!s}".format(previous, current))
        else:
            current = Cellule(value, dt_now)
            # print("{!s} == {!s}".format(previous, current))

        self.dirty = True

        return super().__setitem__(key, (previous, current))

    def learn(self, key, value):
        try:
            self[key] = value
        except CelluleBorked:
            del self[key]
            self[key] = value
        except ValueOutdated:
            return None

        return pendulum.period(self[key][0].dt, self[key][1].dt)

    def current_value(self, key):
        return self[key][-1].value

    def previous_value(self, key):
        return self[key][0].value


class NouNou:
    version = 1
    ENVIRONMENT_BUCKET = "NOUNOU_BUCKET"
    ENVIRONMENT_KEY_DATA_KEY = "NOUNOU_DATA_KEY"

    def __init__(self, data_key=None, *args, **kwargs):
        self.log = logging.getLogger(__name__)
        self.maternelle = dict()
        self.data_key = data_key
        self.bucket = os.environ.get(self.ENVIRONMENT_BUCKET, "poupin")
        self.region_name = kwargs.get("region_name")
        self.changed_items = set()

        if self.data_key is None:
            self.data_key = os.environ.get(self.ENVIRONMENT_KEY_DATA_KEY)

        if self.data_key:
            self.load()

    def load(self, data_key=None):
        self.changed_items = set()
        self.maternelle = dict()

        if data_key:
            self.data_key = data_key

        if self.data_key:
            self.log.debug(
                "Loading data from {!r} using data key {!r}".format(
                    self.bucket, self.data_key
                )
            )
            persistence_control = JSONDocumentOnS3Controller(
                bucket_name=self.bucket, region_name=self.region_name
            )
            raw_data = dict()

            try:
                raw_data = persistence_control[self.data_key]
            except KeyError:
                pass

            try:
                #: transform raw data to instances of ``PoupinCerveau``
                #: and ``Cellule``
                for enfant, enfant_v in raw_data["maternelle"].items():
                    self.maternelle[enfant] = PoupinCerveau()
                    for k, v in enfant_v.items():
                        self.maternelle[enfant][k] = v
            except Exception as exc:
                log_traceback("Failed to transform raw data!", exc, uselog=self.log)
        else:
            self.log.warning("No data key!")

    def persist(self):
        changed_items = copy.deepcopy(self.changed_items)

        if self.data_key:
            for enfant, x in self.maternelle.items():
                if x.dirty:
                    self.log.debug("{!r} is Dirrrty".format(enfant))
                    self.changed_items.add(enfant)
                    changed_items.add(enfant)

            if self.changed_items:
                self.log.info(
                    "Changed: {!s}".format(", ".join(sorted(self.changed_items)))
                )
                persistence_control = JSONDocumentOnS3Controller(
                    bucket_name=self.bucket, region_name=self.region_name
                )
                data = dict(
                    version=self.version,
                    dt=pendulum.now().to_rfc3339_string(),
                    maternelle=self.maternelle,
                )

                persistence_control[self.data_key] = data
                self.changed_items = set()
            else:
                self.log.debug("No persisting needed")
        else:
            self.log.warning("No data key!")

        return changed_items

    def forget(self, enfant, key):
        try:
            del self.maternelle[enfant][key]
            self.changed_items.add(enfant)
        except KeyError:
            pass

    def current_dict(self, enfant):
        current_data = dict()

        for key, data in self.maternelle[enfant].items():
            current_data[key] = data[1].value

        return current_data

    def teach(self, enfant, key, value):
        try:
            self.maternelle[enfant]
        except KeyError:
            self.maternelle[enfant] = PoupinCerveau()

        self.maternelle[enfant].learn(key, value)

    def teach_static(self, enfant, key, value):
        try:
            self.maternelle[enfant]
        except KeyError:
            self.maternelle[enfant] = PoupinCerveau()

        try:
            cv = self.maternelle[enfant].current_value(key)
            if cv != value:
                self.maternelle[enfant].learn(key, value)
        except KeyError:
            self.maternelle[enfant].learn(key, value)


if __name__ == "__main__":
    import doctest

    (FAILED, SUCCEEDED) = doctest.testmod()
    print("[doctest] SUCCEEDED/FAILED: {:d}/{:d}".format(SUCCEEDED, FAILED))
